structure Main =
struct
  structure Seq = ArraySequence
  structure Paren = Paren
  structure Gen = Gen (structure Seq = Seq)
  structure Args = CommandLineArgs

  structure SPM = MkSeqParenMatch (structure Seq = Seq)
  structure SPPM = MkSeqPackedParenMatch (structure Seq = Seq)
  structure PM = MkParenMatch (structure Seq = Seq)
  structure PMG = MkParenMatchG (structure Seq = Seq)
  structure PPMG = MkPackedParenMatchG (structure Seq = Seq)

  (* a useful little tool *)
  fun parensFromString (str : string) : Paren.t Seq.t =
    let
      fun fromChar c =
        case Paren.fromString (Char.toString c) of
          SOME p => p
        | NONE => raise Fail "Inputs must contain only '(' and ')'."
    in
      Seq.map fromChar (Seq.fromList (String.explode str))
    end

  val defaultInput = "simple"
  val defaultSize = 1000
  val defaultBias = 0.5
  val defaultSeed = 15210
  val defaultGran = 1000

  fun simple n =
    Seq.tabulate (fn i => if i < n div 2 then Paren.L else Paren.R) n

  fun getParens () =
    ((*print "Generating input... \n";*)
    case Args.parseOrDefaultString ("input", defaultInput) of
      "simple" => simple (Args.parseOrDefaultInt ("size", defaultSize))
    | "random" => Gen.generate
                    (Args.parseOrDefaultReal ("bias", defaultBias))
                    (Args.parseOrDefaultInt ("size", defaultSize))
                    (Args.parseOrDefaultInt ("seed", defaultSeed))
    | other => parensFromString other)

  val C = Paren.Pack.capacity

  fun packSeq (s : Paren.t Seq.t) : Paren.Pack.t Seq.t =
    let
      fun iSize i = Int.min (C, Seq.length s - (i * C))
      fun iPack i = Paren.Pack.tabulate (fn j => Seq.nth s (i * C + j)) (iSize i)
    in
      Seq.tabulate iPack (1 + (Seq.length s - 1) div C)
    end

  (* assumes every pack is full except the last *)
  fun unpackSeq (s : Paren.Pack.t Seq.t) : Paren.t Seq.t =
    let
      val numParens =
        case Seq.length s of
          0 => 0
        | m => ((m - 1) * C) + (Paren.Pack.length (Seq.nth s (m - 1)))

      fun selectParen i =
        Paren.Pack.nth (Seq.nth s (i div C)) (i mod C)
    in
      Seq.tabulate selectParen numParens
    end

  fun run () =
    let
      val _ = Args.init ()
      val task =
        case Args.parseOrDefaultString ("algo", "ppmg") of
          "spm" =>
            let
              val ps = getParens ()
            in fn () => SPM.parenMatch ps
            end

        | "sppm" =>
            let
              val ps = getParens ()
              val pps = packSeq ps
            in fn () => SPPM.parenMatch pps
            end

        | "pm" =>
            let
              val ps = getParens ()
            in fn () => PM.parenMatch ps
            end

        | "pmg" =>
            let
              val ps = getParens ()
              val g = Args.parseOrDefaultInt ("gran", defaultGran)
            in fn () => PMG.parenMatch g ps
            end

        | "ppmg" =>
            let
              val ps = getParens ()
              val pps = packSeq ps
              val g = Args.parseOrDefaultInt ("gran", defaultGran)
            in fn () => PPMG.parenMatch g pps
            end

        | other => raise Fail ("unknown algo " ^ other)

      (*val _ = print "Running...\n"*)
      val (result, stats) = StatRunner.run task
    in
      print (String.concat ["result ", Bool.toString result, "\n",
                            stats])
    end
end

val () = Main.run ()
