structure Paren =
struct
  datatype t = L | R
  type paren = t

  fun fromString str =
    if String.size str <> 1 then NONE else
    case String.sub (str, 0) of
      #"(" => SOME L
    | #")" => SOME R
    | _ => NONE

  fun toString p =
    case p of
      L => "("
    | R => ")"

  structure Pack :>
  sig
    type t
    val capacity : int
    val length : t -> int
    val nth : t -> int -> paren
    val tabulate : (int -> paren) -> int -> t
    val iterate : ('a * paren -> 'a) -> 'a -> t -> 'a
  end =
  struct
    structure W = LargeWord

    (* A paren pack can contain at most `capacity` parens. We reserve `lbits` least
     * significant bits to store the length. The remaining `capacity` bits are
     * either 0 for L and 1 for R, stored in order from least
     * significant to most significant. *)
    type t = W.word

    (* Adjust how many bits are used to store the length until there are enough
     * so that 2^lbits - 1 >= capacity. The goal is to be able to construct a
     * ParenPack for any wordsize (LargeWords do not have a specified wordsize),
     * but still maximize the bit usage of a single word. All of this happens at
     * compile time. *)
    fun determineCapacity cap maxLen =
      if maxLen >= cap then cap else determineCapacity (cap-1) (2*(maxLen+1) - 1)

    val capacity : int = determineCapacity (W.wordSize-1) 1
    val lbits : int = W.wordSize - capacity

    val wzero = W.fromInt 0
    val wone = W.fromInt 1

    val wfi = Word.fromInt

    val lengthMask = W.- (W.<< (wone, wfi lbits), wone)
    fun length ps = W.toInt (W.andb (ps, lengthMask))

    fun nth (ps : t) i =
      if i < 0 orelse i > length ps then raise Fail ("Attempting to extract " ^ Int.toString i ^ "th paren from pack of size " ^ Int.toString (length ps)) else
      if W.andb (W.>> (ps, wfi lbits + wfi i), wone) = wzero then L else R

    fun tabulate f n =
      if n < 0 orelse n > capacity then raise Fail ("Parenthesis pack capacity is " ^ Int.toString capacity ^ ", cannot create one of size " ^ Int.toString n) else
      let
        fun loop ps i =
          if i = n then ps else
          case f i of
            L => loop ps (i+1) (* ps is initialized to all 0's *)
          | R => loop (W.orb (ps, W.<< (wone, wfi lbits + wfi i))) (i+1)
      in
        loop (W.fromInt n) 0
      end

    fun iterate f b ps =
      let
        val n = length ps
        fun loop x ps i =
          if i = n then x else
          let val x' = f (x, if W.andb (ps, wone) = wzero then L else R)
              val ps' = W.>> (ps, 0w1)
          in loop x' ps' (i+1)
          end
      in
        loop b (W.>> (ps, wfi lbits)) 0
      end
  end (* Paren.Pack *)

end
